package info.textgrid.middleware;

import static org.junit.Assert.assertTrue;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.ws.rs.core.Response;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.Client;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.http.HttpStatus;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

/**
 * <p>
 * Some online tests for the TextGrid OAIMPH service --> verb=GetRecord.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 * @version 2022-09-08
 * @since 2022-09-08
 */
@Ignore
public class TestGetRecordOnline {

  // **
  // FINALS
  // **

  // private static final String PROPERTIES_FILE = "oaipmh.test.textgridlab-org.properties";
  // TODO There is NO IDIOM test data for testing on dev.textgridlab.org! So eventually all
  // IDIOM tests will fail!
  private static final String PROPERTIES_FILE = "oaipmh.test.dev-textgridlab-org.properties";

  // **
  // STATICS
  // **

  // Some JAXRS things.
  private static String oaipmhEndpoint;
  private static OAIPMHProducer JAXRSClient;
  static Client oaipmhWebClient;

  // Properties
  private static String checkGetRecordDC;
  private static String expectedGetRecordDC;
  private static String checkGetRecordIDIOM;
  private static String checkGetRecordIDIOM_Images;
  private static String expectedGetRecordIDIOM_Images;
  private static String expectedGetRecordIDIOM;
  private static String checkGetRecordDATACITE;
  private static String expectedGetRecordDATACITE;
  private static List<String> checkGetRecordIDList = new ArrayList<String>();

  // **
  // PREPARATIONS
  // **

  /**
   * @throws Exception
   */
  @BeforeClass
  public static void setUpBeforeClass() throws Exception {

    // Load properties file.
    Properties p = new Properties();
    p.load(new FileInputStream(OAIPMHUtilitiesOnline.getResource(PROPERTIES_FILE)));

    System.out.println("Properties file: " + PROPERTIES_FILE);
    p.list(System.out);

    // Get other needed properties.
    oaipmhEndpoint = p.getProperty("oaipmhEndpoint");
    checkGetRecordDC = p.getProperty("checkGetRecordDC");
    expectedGetRecordDC = p.getProperty("expectedGetRecordDC");
    checkGetRecordIDIOM = p.getProperty("checkGetRecordIDIOM");
    checkGetRecordIDIOM_Images = p.getProperty("checkGetRecordIDIOM_Images");
    expectedGetRecordIDIOM_Images = p.getProperty("expectedGetRecordIDIOM_Images");
    expectedGetRecordIDIOM = p.getProperty("expectedGetRecordIDIOM");
    checkGetRecordDATACITE = p.getProperty("checkGetRecordDATACITE");
    expectedGetRecordDATACITE = p.getProperty("expectedGetRecordDATACITE");
    checkGetRecordIDList =
        OAIPMHUtilitiesOnline.getListFromProperties((String) p.get("checkGetRecordIDList"));

    // Get OAI-PMH REST endpoint and HTTP client.
    System.out.println("Getting OAI-PMH HTTP client --> " + oaipmhEndpoint + " <--");

    // Get proxy first, set policy.
    JAXRSClient = JAXRSClientFactory.create(oaipmhEndpoint, OAIPMHProducer.class);
    HTTPConduit conduit = WebClient.getConfig(JAXRSClient).getHttpConduit();
    HTTPClientPolicy policy = new HTTPClientPolicy();
    policy.setReceiveTimeout(OAIPMHUtilitiesOnline.OAIPMH_CLIENT_TIMEOUT);
    conduit.setClient(policy);

    // Create Web Client from Web Proxy.
    oaipmhWebClient = WebClient.client(JAXRSClient);
  }

  /**
   * @throws Exception
   */
  @AfterClass
  public static void tearDownAfterClass() throws Exception {
    //
  }

  /**
   * @throws Exception
   */
  @Before
  public void setUp() throws Exception {
    //
  }

  /**
   * @throws Exception
   */
  @After
  public void tearDown() throws Exception {
    //
  }

  // **
  // TESTS
  // **

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDC() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordOAIDC()");

    String path = "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDC + "&metadataPrefix="
        + OAIPMHUtilitiesOnline.OAI_DC_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDC)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordDC + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDCNoPrefix() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordOAIDCNoPrefix()");

    String path = "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDC + "&metadataPrefix="
        + OAIPMHUtilitiesOnline.OAI_DC_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDC)) {
      System.err.println("\tstatus: " + status);
      System.err.println(
          OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
              + checkGetRecordDC + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordIDIOMMETS()");

    String path = "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
        + checkGetRecordIDIOM + "&metadataPrefix="
        + OAIPMHUtilitiesOnline.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS_Image() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordIDIOMMETS_Image()");

    String path =
        "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
            + checkGetRecordIDIOM_Images
            + "&metadataPrefix=" + OAIPMHUtilitiesOnline.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM_Images)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM_Images + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETSNoPrefix() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordIDIOMMETSNoPrefix()");

    String path = "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
        + checkGetRecordIDIOM + "&metadataPrefix="
        + OAIPMHUtilitiesOnline.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordIDIOM)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordIDIOM + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITE() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordDATACITE()");

    String path = "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
        + checkGetRecordDATACITE + "&metadataPrefix="
        + OAIPMHUtilitiesOnline.OAI_DATACITE_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDATACITE)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordDATACITE + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITENoPrefix() throws IOException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testGetRecordDATACITENoPrefix()");

    String path =
        "verb=" + OAIPMHUtilitiesOnline.VERB_GET_RECORD + "&identifier="
            + checkGetRecordDATACITE
            + "&metadataPrefix=" + OAIPMHUtilitiesOnline.OAI_DATACITE_PREFIX;

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, path,
        OAIPMHUtilitiesOnline.NO_THREAD_NAME);
    int status = httpResponse.getStatus();
    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK
        || !response.contains(expectedGetRecordDATACITE)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OAIPMHUtilitiesOnline.ERROR + ": response should contain '"
          + expectedGetRecordDATACITE + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OAIPMHUtilitiesOnline.OK);
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testConcurrentlyOAIDCGetRecord() throws InterruptedException, ExecutionException {

    System.out.println(OAIPMHUtilitiesOnline.TESTING + "testConcurrentlyOAIDCGetRecord()");

    ExecutorService executor = Executors.newFixedThreadPool(18);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(0),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(1),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(2),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f4 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(3),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f5 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(4),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f6 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(5),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f7 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(6),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f8 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(7),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f9 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(8),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f10 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(9),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f11 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(10),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f12 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(11),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f13 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(12),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f14 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(13),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f15 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(14),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f16 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(15),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f17 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(16),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));
    Future<Boolean> f18 = executor
        .submit(new OaiPmhGetRecordTestThread(OAIPMHUtilitiesOnline.VERB_GET_RECORD,
            checkGetRecordIDList.get(17),
            OAIPMHUtilitiesOnline.OAI_DC_PREFIX,
            "GR[" + checkGetRecordIDList.get(0) + "]"));

    executor.shutdown();

    System.out.println(OAIPMHUtilitiesOnline.OK + "[GR1]=" + f1.get() + ", [GR2]=" + f2.get()
        + ", [GR3]=" + f3.get() + ", [GR4]=" + f4.get() + ", [GR5]=" + f5.get() + ", [GR6]="
        + f6.get() + ", [GR7]=" + f7.get() + ", [GR8]=" + f8.get() + ", [GR9]=" + f9.get()
        + ", [GR10]=" + f10.get() + ", [GR11]=" + f11.get() + ", [GR12]=" + f12.get()
        + ", [GR13]=" + f13.get() + ", [GR14]=" + f14.get() + ", [GR15]=" + f15.get()
        + ", [GR16]=" + f16.get() + ", [GR17]=" + f17.get() + ", [GR18]=" + f18.get());
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */
  private class OaiPmhGetRecordTestThread implements Callable<Boolean> {

    protected String verb;
    protected String identifier;
    protected String metadataPrefix;
    protected String threadName;

    /**
     * @param theVerb
     * @param theIdentifier
     * @param theMetadataPrefix
     * @param theThreadName
     */
    public OaiPmhGetRecordTestThread(String theVerb, String theIdentifier, String theMetadataPrefix,
        String theThreadName) {
      this.verb = theVerb;
      this.identifier = theIdentifier;
      this.metadataPrefix = theMetadataPrefix;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.concurrent.Callable#call()
     */
    @Override
    public Boolean call() throws IOException {

      String url = "verb=" + this.verb + "&identifier=" + this.identifier + "&metadataPrefix="
          + this.metadataPrefix;
      OAIPMHUtilitiesOnline.getHttpResponse(oaipmhWebClient, url, this.threadName);

      return true;
    }
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theClient
   * @param theVerb
   * @param theSet
   * @param theMetadataPrefix
   * @param maxNumberOfPagesToTest
   * @param recordsExpectedPerRequest
   * @param theThreadName
   * @param from
   * @param until
   * @param resumptionTokenANDMetadataPrefix
   * @throws IOException
   */
  protected static void testList(Client theClient, final String theVerb, final String theSet,
      final String theMetadataPrefix, final int maxNumberOfPagesToTest,
      final int recordsExpectedPerRequest, final String theThreadName, final String from,
      final String until, final boolean resumptionTokenANDMetadataPrefix) throws IOException {

    String threadName = tn(theThreadName);

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(OAIPMHUtilitiesOnline.VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String path = "verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null) {
      path += "&set=" + theSet;
    }

    if (from != null && until != null) {
      path += "&from=" + from + "&until=" + until;
    }

    Response httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(theClient, path, threadName);
    int status = httpResponse.getStatus();

    int loopCount = 1;
    long timeRunning = System.currentTimeMillis() - startTime;
    System.out.println("\t" + threadName + "time: "
        + OAIPMHUtilitiesOnline.getDurationInSecs(timeRunning) + " (loop " + loopCount
        + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

    String responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    // Test resumption token tags.
    String restok = examineResumptionTokenTag(responseString, testOccurance,
        OAIPMHUtilitiesOnline.NO_TOKEN, recordsExpectedPerRequest, loopCount, threadName);

    // Test general metadata content (if verb is listRecords!), must go conform with the metadata
    // prefix setting.
    if (theVerb.equals(OAIPMHUtilitiesOnline.VERB_LIST_RECORDS)) {
      examineContent(responseString, theMetadataPrefix);
    }

    while (status == HttpStatus.SC_OK && !restok.equals(OAIPMHUtilitiesOnline.NO_TOKEN)) {
      loopCount += 1;
      synchronized (threadName) {
        if (maxNumberOfPagesToTest > 0 && loopCount > maxNumberOfPagesToTest) {
          System.out
              .println("\t" + threadName + "TESTING ONLY " + maxNumberOfPagesToTest + " pages!");
          break;
        }
        path = "verb=" + theVerb + "&resumptionToken=" + restok;
        if (resumptionTokenANDMetadataPrefix) {
          path += "&metadataPrefix=" + theMetadataPrefix;
        }
        httpResponse = OAIPMHUtilitiesOnline.getHttpResponse(theClient, path, threadName);
        timeRunning = System.currentTimeMillis() - startTime;
        System.out.println("\t" + threadName + "time: "
            + OAIPMHUtilitiesOnline.getDurationInSecs(timeRunning) + " (loop " + loopCount
            + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

        // Test resumption token tags.
        responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());
        restok = examineResumptionTokenTag(responseString, testOccurance, restok,
            recordsExpectedPerRequest, loopCount, threadName);

        // Test general metadata content (if verb is listRecords!), must go conform with the
        // metadata // prefix setting.
        if (theVerb.equals(OAIPMHUtilitiesOnline.VERB_LIST_RECORDS)) {
          examineContent(responseString, theMetadataPrefix);
        }
      }
    }

    // Only check for max loops, if we do have less loops and no resumption token, it will be fine,
    // too!
    if (!restok.equals(OAIPMHUtilitiesOnline.NO_TOKEN) && loopCount < maxNumberOfPagesToTest) {
      System.out.println(threadName + OAIPMHUtilitiesOnline.ERROR + ": Must have done "
          + maxNumberOfPagesToTest + " loops, but did only " + loopCount);
      assertTrue(false);
    }

    // Check for single loop, we need MORE!
    if (loopCount == 1) {
      System.out
          .println(
              threadName + OAIPMHUtilitiesOnline.ERROR + ": Must have more than one loop!");
      assertTrue(false);
    }

    System.out.println("\t" + threadName + OAIPMHUtilitiesOnline.OK);
  }

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theResponse
   * @param recordOrHeader
   * @param oldtok
   * @param recordsExpectedPerRequest
   * @param loopCount
   * @param theThreadName
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  public static String examineResumptionTokenTag(final String theResponseString,
      final String recordOrHeader, final String oldtok, final int recordsExpectedPerRequest,
      final int loopCount, final String theThreadName) throws IOException {

    // Test for OAIPMH errors.
    if (theResponseString.contains("<error code=\"badArgument\">")) {
      String message =
          theThreadName + OAIPMHUtilitiesOnline.ERROR + " IN OAIPMH RESPONSE: "
              + theResponseString;
      System.out.println(message);
      throw new IOException(message);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = theResponseString.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = theResponseString.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + theThreadName + recordOrHeader + "s: " + recordCount);

    // Check if token tag is existing.
    int tokStart = theResponseString.indexOf("<resumptionToken");
    int tokEnd = theResponseString.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\t" + theThreadName + "token: no token");

      return OAIPMHUtilitiesOnline.NO_TOKEN;
    }

    String restokTmp = theResponseString.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "tokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "token: " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\t" + theThreadName + "tokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\t" + theThreadName + "size: " + size + " / " + cursor);

    // If token is provided, test cursor and element count!
    if (!restok.isEmpty()) {
      synchronized (TestGetRecordOnline.class) {
        // Check <record> or <header> count, must be recordsExpectedPerRequest!
        if (recordCount != recordsExpectedPerRequest) {
          String message =
              OAIPMHUtilitiesOnline.ERROR + ": " + recordOrHeader + " count mismatch, must be "
                  + recordsExpectedPerRequest + " if token is provided, but is " + recordCount;
          assertTrue(message, false);
        }
        if (size <= recordsExpectedPerRequest) {
          String message =
              OAIPMHUtilitiesOnline.ERROR + ": completeListSize count mismatch, must be > "
                  + recordsExpectedPerRequest + " if token is provided, but is " + size;
          assertTrue(message, false);
        }
        if (cursor != recordsExpectedPerRequest * loopCount) {
          String message = OAIPMHUtilitiesOnline.ERROR + ": cursor must be "
              + (loopCount * recordsExpectedPerRequest) + " in loop " + loopCount + ", but is "
              + cursor;
          assertTrue(message, false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % recordsExpectedPerRequest.
      if (recordCount != size % recordsExpectedPerRequest) {
        System.err.println(theThreadName + OAIPMHUtilitiesOnline.ERROR + ": " + recordOrHeader
            + " count mismatch, should be " + size % recordsExpectedPerRequest + ", but is "
            + recordCount);
        assertTrue(false);
      }

      // No resumption token available in response.
      return OAIPMHUtilitiesOnline.NO_TOKEN;
    }

    return restok;
  }

  /**
   * @param theThreadName
   * @return
   */
  private static String tn(String theThreadName) {

    String result = theThreadName;

    if (theThreadName != null && !theThreadName.equals("") && !theThreadName.endsWith(" ")) {
      result = "[" + theThreadName + "] ";
    }

    return result;
  }

  /**
   * @param theResponseString
   * @param theMetadataFormat
   */
  private static void examineContent(String theResponseString, String theMetadataFormat) {

    // Check for correct metadata content according to metadata prefix.
    if (theMetadataFormat.equals(OAIPMHUtilitiesOnline.OAI_DC_PREFIX)
        && !theResponseString
            .contains("metadataPrefix=\"" + OAIPMHUtilitiesOnline.OAI_DC_PREFIX + "\"")
        && !theResponseString.contains(OAIPMHUtilitiesOnline.EXPECTED_OAIDC_FORMAT_CONTENT)) {
      System.out
          .println(
              OAIPMHUtilitiesOnline.OAI_DC_PREFIX + " needs to deliver content with schema: "
                  + OAIPMHUtilitiesOnline.EXPECTED_OAIDC_FORMAT_CONTENT + "!");

      System.out.println("ERROR:\n" + theResponseString);

      assertTrue(false);
    } else if (theMetadataFormat.equals(OAIPMHUtilitiesOnline.OAI_IDIOMMETS_PREFIX)
        && !theResponseString.contains(
            "metadataPrefix=\"" + OAIPMHUtilitiesOnline.EXPECTED_IDIOMMETS_FORMAT_CONTENT
                + "\"")
        && !theResponseString
            .contains(OAIPMHUtilitiesOnline.EXPECTED_IDIOMMETS_FORMAT_CONTENT)) {
      System.out.println(
          OAIPMHUtilitiesOnline.OAI_IDIOMMETS_PREFIX + " needs to deliver content with schema: "
              + OAIPMHUtilitiesOnline.EXPECTED_IDIOMMETS_FORMAT_CONTENT + "!");

      System.out.println("ERROR:\n" + theResponseString);

      assertTrue(false);
    } else if (theMetadataFormat.equals(OAIPMHUtilitiesOnline.OAI_DATACITE_PREFIX)
        && !theResponseString
            .contains("metadataPrefix=\"" + OAIPMHUtilitiesOnline.OAI_DATACITE_PREFIX + "\"")
        && !theResponseString
            .contains(OAIPMHUtilitiesOnline.EXPECTED_DATACITE_FORMAT_CONTENT)) {
      System.out.println(
          OAIPMHUtilitiesOnline.OAI_DATACITE_PREFIX + " needs to deliver content with schema: "
              + OAIPMHUtilitiesOnline.EXPECTED_DATACITE_FORMAT_CONTENT + "!");

      System.out.println("ERROR:\n" + theResponseString);

      assertTrue(false);
    } else {
      System.out.println(theResponseString);
    }
  }

}
